@@ -5,12 +5,13 @@ from p2p_python.tool.traffic import Traffic
 from p2p_python.tool.utils import AESCipher
 from ecdsa.keys import SigningKey, VerifyingKey
 from ecdsa.curves import NIST256p
-from typing import Optional, Union, List
+from typing import Optional, Dict, List
 from logging import getLogger
 from binascii import a2b_hex
 from time import time
 from io import BytesIO
 from hashlib import sha256
+from expiringdict import ExpiringDict
 from asyncio.streams import StreamWriter, StreamReader
 import asyncio
 import json
@@ -20,9 +21,9 @@ import socks
 import zlib
 
 
-# constant
-SERVER_SIDE = 'server'
-CLIENT_SIDE = 'client'
+# socket direction
+INBOUND = 'inbound'
+OUTBOUND = 'outbound'
 
 log = getLogger(__name__)
 loop = asyncio.get_event_loop()
@@ -55,6 +56,7 @@ class Core(object):
         self.core_que = asyncio.Queue()
         self.backlog = listen
         self.traffic = Traffic()
+        self.ping_status: Dict[int, asyncio.Event] = ExpiringDict(max_len=5000, max_age_seconds=900)
 
     def close(self):
         if not self.f_running:
@@ -71,24 +73,25 @@ class Core(object):
         self.f_stop = True
 
     async def ping(self, user: User, f_udp=False):
+        uuid = random.randint(1000000000, 4294967295)
         try:
-            await asyncio.wait_for(user.event.wait(), 10.0)
-            user.event.clear()
-            await self.send_msg_body(msg_body=b'Ping', user=user, allow_udp=f_udp, f_pro_force=True)
-            await asyncio.wait_for(user.event.wait(), 5.0)
-            user.event.set()
+            # prepare Event
+            event = asyncio.Event()
+            self.ping_status[uuid] = event
+            # send ping
+            msg_body = b'Ping:' + str(uuid).encode()
+            await self.send_msg_body(msg_body=msg_body, user=user, allow_udp=f_udp, f_pro_force=True)
+            # wait for event set (5s)
+            await asyncio.wait_for(event.wait(), 5.0)
             return True
         except asyncio.TimeoutError:
             log.debug(f"failed to udp ping {user}")
-            user.event.set()
-            return False
         except ConnectionError as e:
             log.debug(f"socket error on ping by {e}")
-            user.event.set()
-            return False
         except Exception:
             log.error("ping exception", exc_info=True)
-            return False
+        # failed
+        return False
 
     def start(self, s_family=socket.AF_UNSPEC):
         assert s_family in (socket.AF_INET, socket.AF_INET6, socket.AF_UNSPEC)
@@ -115,8 +118,14 @@ class Core(object):
     async def create_connection(self, host, port):
         if self.f_stop:
             return False
-        for res in socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM):
-            af, socktype, proto, canonname, host_port = res
+        # get connection list
+        future = loop.run_in_executor(
+            None, socket.getaddrinfo, host, port, socket.AF_UNSPEC, socket.SOCK_STREAM)
+        await asyncio.wait_for(future, 10.0)
+        if future.cancelled():
+            return False
+        # try to connect one by one
+        for af, socktype, proto, canonname, host_port in future.result():
             if host_port[0] in ban_address:
                 return False  # baned address
             try:
@@ -127,7 +136,6 @@ class Core(object):
                     sock.setproxy(socks.PROXY_TYPE_SOCKS5, V.TOR_CONNECTION[0], V.TOR_CONNECTION[1])
                 else:
                     sock = socket.socket(af, socktype, proto)
-                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                 future = loop.run_in_executor(None, sock.connect, host_port)
                 await asyncio.wait_for(future, 10.0)
                 future.result()  # raised exception of socket
@@ -193,7 +201,7 @@ class Core(object):
 
             # 6. generate new user
             user_header = UserHeader(**header)
-            new_user = User(user_header, self.number, reader, writer, host_port, aeskey, CLIENT_SIDE)
+            new_user = User(user_header, self.number, reader, writer, host_port, aeskey, OUTBOUND)
 
             # 7. check header
             if new_user.header.network_ver != V.NETWORK_VER:
@@ -210,8 +218,7 @@ class Core(object):
             log.info(f"established connection as client to {new_user.header.name} {new_user.get_host_port()}")
             asyncio.ensure_future(self.receive_loop(new_user))
             # server port's reachable check
-            await asyncio.sleep(1.0)
-            await self.check_reachable(new_user)
+            asyncio.ensure_future(self.check_reachable(new_user))
             return True
         except PeerToPeerError as e:
             msg = "peer2peer error, {} ({})".format(e, host)
@@ -236,9 +243,9 @@ class Core(object):
         if user in self.user:
             self.user.remove(user)
             if 0 < user.score:
-                log.info(f"remove connection of {user.header.name} by '{reason}'")
+                log.info(f"remove connection of {user} by '{reason}'")
             else:
-                log.debug(f"remove connection of {user.header.name} by '{reason}'")
+                log.debug(f"remove connection of {user} by '{reason}'")
             return True
         else:
             return False
@@ -300,7 +307,7 @@ class Core(object):
 
             # 3. generate new user
             user_header = UserHeader(**header)
-            new_user = User(user_header, self.number, reader, writer, host_port, AESCipher.create_key(), SERVER_SIDE)
+            new_user = User(user_header, self.number, reader, writer, host_port, AESCipher.create_key(), INBOUND)
             self.number += 1
             if new_user.header.name == V.SERVER_NAME:
                 raise ConnectionAbortedError('Same origin connection')
@@ -347,8 +354,7 @@ class Core(object):
             log.info(f"established connection as server from {new_user.header.name} {new_user.get_host_port()}")
             asyncio.ensure_future(self.receive_loop(new_user))
             # server port's reachable check
-            await asyncio.sleep(1.0)
-            await self.check_reachable(new_user)
+            asyncio.ensure_future(self.check_reachable(new_user))
             return
         except (ConnectionAbortedError, ConnectionResetError) as e:
             msg = f"disconnect error {host_port} {e}"
@@ -446,14 +452,17 @@ class Core(object):
                 self.traffic.put_traffic_down(msg_body)
                 msg_body = AESCipher.decrypt(key=user.aeskey, enc=msg_body)
                 msg_body = zlib.decompress(msg_body)
-                if msg_body == b'Ping':
+                if msg_body.startswith(b'Ping:'):
+                    uuid_bytes = msg_body.split(b':')[1]
                     log.debug(f"receive Ping from {user.header.name}")
-                    await self.send_msg_body(b'Pong', user)
-                elif msg_body == b'Pong':
-                    log.debug(f"receive Pong from {user.header.name}")
-                    user.event.set()
+                    await self.send_msg_body(b'Pong:' + uuid_bytes, user)
+                elif msg_body.startswith(b'Pong:'):
+                    uuid_int = int(msg_body.decode().split(':')[1])
+                    if uuid_int in self.ping_status:
+                        log.debug(f"receive Pong from {user.header.name}")
+                        self.ping_status[uuid_int].set()
                 else:
-                    await self.core_que.put((user, msg_body))
+                    await self.core_que.put((user, msg_body, time()))
                 f_raise_timeout = False
 
             except asyncio.TimeoutError:
@@ -485,30 +494,31 @@ class Core(object):
                     log.debug(f"user connection closed on check_reachable {new_user}")
                     return
                 await asyncio.sleep(1.0)
-            # try to check
+            # try to check TCP
             f_tcp = True
             host_port = new_user.get_host_port()
             af = socket.AF_INET if len(host_port) == 2 else socket.AF_INET6
             sock = socket.socket(af, socket.SOCK_STREAM)
-            sock.settimeout(5.0)
+            sock.settimeout(3.0)
             try:
                 future = loop.run_in_executor(None, sock.connect_ex, host_port)
-                await asyncio.wait_for(future, 60.0)
+                await asyncio.wait_for(future, 10.0)
                 result = future.result()
                 if result != 0:
                     f_tcp = False
-                sock.close()
             except OSError:
                 f_tcp = False
+            loop.run_in_executor(None, sock.close)
+            # try to check UDP
             f_udp = await self.ping(user=new_user, f_udp=True)
             f_changed = False
             # reflect user status
             if f_tcp is not new_user.header.p2p_accept:
-                log.debug(f"{new_user} Update TCP accept status {new_user.header.p2p_accept}=>{f_tcp}")
+                # log.debug(f"{new_user} Update TCP accept status {new_user.header.p2p_accept}=>{f_tcp}")
                 new_user.header.p2p_accept = f_tcp
                 f_changed = True
             if f_udp is not new_user.header.p2p_udp_accept:
-                log.debug(f"{new_user} Update UDP accept status {new_user.header.p2p_udp_accept}=>{f_udp}")
+                # log.debug(f"{new_user} Update UDP accept status {new_user.header.p2p_udp_accept}=>{f_udp}")
                 new_user.header.p2p_udp_accept = f_udp
                 f_changed = True
             if f_changed:
@@ -516,6 +526,25 @@ class Core(object):
         except Exception:
             log.error("check_reachable exception", exc_info=True)
 
+    async def try_reconnect(self, user: User, reason: str):
+        self.remove_connection(user, reason)
+        host_port = user.get_host_port()
+        if self.f_stop:
+            return False
+        elif not user.header.p2p_accept:
+            return False
+        elif await self.create_connection(host=host_port[0], port=host_port[1]):
+            log.debug(f"reconnect success {user}")
+            new_user = self.host_port2user(host_port)
+            if new_user:
+                new_user.neers = user.neers
+                new_user.score = user.score
+                new_user.warn = user.warn
+            return True
+        else:
+            log.warning(f"reconnect failed {user}")
+            return False
+
     def name2user(self, name) -> Optional[User]:
         for user in self.user:
             if user.header.name == name:
@@ -559,12 +588,13 @@ async def udp_server_handle(msg, addr, core: Core):
             return
         core.traffic.put_traffic_down(msg_body)
         msg_body = AESCipher.decrypt(key=user.aeskey, enc=msg_body)
-        if msg_body == b'Ping':
-            log.info(f"get udp ping from {user.header.name} addr:{addr}")
-            await core.send_msg_body(msg_body=b'Pong', user=user)
+        if msg_body.startswith(b'Ping:'):
+            log.info(f"get udp ping from {user}")
+            uuid_bytes = msg_body.split(b':')[1]
+            await core.send_msg_body(msg_body=b'Pong:' + uuid_bytes, user=user)
         else:
-            log.debug(f"get udp packet from {user.header.name} addr:{addr}")
-            await core.core_que.put((user, msg_body))
+            log.debug(f"get udp packet from {user}")
+            await core.core_que.put((user, msg_body, time()))
     except ValueError as e:
         log.debug(f"maybe decrypt failed by {e} {msg_body}")
     except OSError as e:
@@ -578,7 +608,10 @@ def create_tcp_server(core: Core, family, host_port):
     coroutine = asyncio.start_server(
         core.initial_connection_check, host_port[0], host_port[1],
         family=family, backlog=core.backlog, loop=loop)
-    return loop.run_until_complete(coroutine)
+    abstract_server = loop.run_until_complete(coroutine)
+    for sock in abstract_server.sockets:
+        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+    return abstract_server
 
 
 def create_udp_server(core: Core, family, host_port):
@@ -631,8 +664,8 @@ def setup_all_socket_server(core: Core, s_family):
 
 
 __all__ = [
-    "SERVER_SIDE",
-    "CLIENT_SIDE",
+    "INBOUND",
+    "OUTBOUND",
     "ban_address",
     "Core",
 ]
